import tensorflow as tf

from spektral.layers import ops
from spektral.layers.convolutional.conv import Conv
from spektral.utils.convolution import gcn_filter, incidence_matrix, line_graph


class CensNetConv(Conv):
    r"""
    A CensNet convolutional layer from the paper

    > [Co-embedding of Nodes and Edges with Graph Neural Networks](https://arxiv.org/abs/2010.13242)<br>
    > Xiaodong Jiang et al.

    This implements both the node and edge propagation rules as a single layer.

    **Mode**: single, disjoint, batch.

    **Input**

    - Node features of shape `([batch], n_nodes, n_node_features)`;
    - A tuple containing:
        - Modified Laplacian of shape `([batch], n_nodes, n_nodes)`; can be
            computed with `spektral.utils.convolution.gcn_filter`.
        - Modified line graph Laplacian of shape `([batch], n_edges, n_edges)`;
            can be computed with `spektral.utils.convolution.line_graph` and
            `spektral.utils.convolution.gcn_filter`.
        - Incidence matrix of shape `([batch], n_nodes, n_edges)`; can be
            computed with `spektral.utils.convolution.incidence_matrix`.
    - Edge features of shape `([batch], n_edges, n_edge_features)`;

    **Output**

    - Node features with the same shape as the input, but with the last
    dimension changed to `node_channels`.
    - Edge features with the same shape as the input, but with the last
    dimension changed to `edge_channels`.

    **Arguments**

    - `node_channels`: number of output channels for the node features;
    - `edge_channels`: number of output channels for the edge features;
    - `activation`: activation function;
    - `use_bias`: bool, add a bias vector to the output;
    - `kernel_initializer`: initializer for the weights;
    - `node_initializer`: initializer for the node feature weights (P_n);
    - `edge_initializer`: initializer for the edge feature weights (P_e);
    - `bias_initializer`: initializer for the bias vector;
    - `kernel_regularizer`: regularization applied to the weights;
    - `edge_regularizer`: regularization applied to the edge feature weights;
    - `bias_regularizer`: regularization applied to the bias vector;
    - `activity_regularizer`: regularization applied to the output;
    - `kernel_constraint`: constraint applied to the weights;
    - `edge_constraint`: constraint applied to the edge feature weights;
    - `bias_constraint`: constraint applied to the bias vector.
    """

    def __init__(
        self,
        node_channels,
        edge_channels,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        node_initializer="glorot_uniform",
        edge_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        node_regularizer=None,
        edge_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        node_constraint=None,
        edge_constraint=None,
        bias_constraint=None,
        **kwargs
    ):
        super().__init__(
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )

        self.node_channels = node_channels
        self.edge_channels = edge_channels

        self.__node_initializer = tf.keras.initializers.get(node_initializer)
        self.__node_regularizer = tf.keras.regularizers.get(node_regularizer)
        self.__node_constraint = tf.keras.constraints.get(node_constraint)

        self.__edge_initializer = tf.keras.initializers.get(edge_initializer)
        self.__edge_regularizer = tf.keras.regularizers.get(edge_regularizer)
        self.__edge_constraint = tf.keras.constraints.get(edge_constraint)

    def build(self, input_shape):
        assert len(input_shape) >= 2
        node_features_shape, _, edge_features_shape = input_shape
        num_input_node_features = node_features_shape[-1]
        num_input_edge_features = edge_features_shape[-1]

        self.node_kernel = self.add_weight(
            shape=(num_input_node_features, self.node_channels),
            initializer=self.kernel_initializer,
            name="node_kernel",
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
        )
        self.edge_kernel = self.add_weight(
            shape=(num_input_edge_features, self.edge_channels),
            initializer=self.kernel_initializer,
            name="edge_kernel",
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
        )

        # Add separate weight vectors for the edge features for the node and
        # edge feature vectors. (These are P_n and P_e in the paper.)
        self.node_weights = self.add_weight(
            shape=(num_input_node_features, 1),
            initializer=self.__node_initializer,
            name="node_weights",
            regularizer=self.__node_regularizer,
            constraint=self.__node_constraint,
        )
        self.edge_weights = self.add_weight(
            shape=(num_input_edge_features, 1),
            initializer=self.__edge_initializer,
            name="edge_weights",
            regularizer=self.__edge_regularizer,
            constraint=self.__edge_constraint,
        )

        if self.use_bias:
            self.node_bias = self.add_weight(
                shape=(self.node_channels,),
                initializer=self.bias_initializer,
                name="node_bias",
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
            )
            self.edge_bias = self.add_weight(
                shape=(self.edge_channels,),
                initializer=self.bias_initializer,
                name="edge_bias",
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
            )
        self.built = True

    def _bias_and_activation(self, pre_activation, *, bias_weights, mask=None):
        """
        Applies the bias, activation, and mask, if necessary.

        :param pre_activation: The layer output, pre-activation.
        :param bias_weights: The weights to use for the bias.
        :param mask: The mask to use.
        :return: The biased, activated, and masked output.
        """

        if self.use_bias:
            # Apply the bias if needed.
            pre_activation = tf.nn.bias_add(pre_activation, bias_weights)
        if mask is not None:
            pre_activation *= mask[0]
        return self.activation(pre_activation)

    def _propagate_nodes(self, inputs, mask=None):
        """
        Performs the node feature propagation step.

        :param inputs: All the inputs to the layer.
        :param mask: The mask to use.
        :return: The propagated node features.
        """
        node_features, (laplacian, _, incidence), edge_features = inputs

        weighted_edge_features = tf.matmul(edge_features, self.edge_weights)
        # Remove the extra 1-dimension.
        weighted_edge_features = tf.squeeze(weighted_edge_features, axis=[-1])
        weighted_edge_features = tf.linalg.diag(weighted_edge_features)
        weighted_edge_features = ops.modal_dot(incidence, weighted_edge_features)
        weighted_edge_features = ops.modal_dot(
            weighted_edge_features, incidence, transpose_b=True
        )

        node_adjacency = weighted_edge_features * laplacian
        output = ops.modal_dot(node_adjacency, node_features)
        output = ops.modal_dot(output, self.node_kernel)

        return self._bias_and_activation(output, bias_weights=self.node_bias, mask=mask)

    def _propagate_edges(self, inputs, mask=None):
        """
        Performs the edge feature propagation step.

        :param inputs: All the inputs to the layer.
        :param mask: The mask to use.
        :return: The propagated edge features.
        """
        node_features, (_, laplacian, incidence), edge_features = inputs

        weighted_node_features = tf.matmul(node_features, self.node_weights)
        # Remove the extra 1-dimension.
        weighted_node_features = tf.squeeze(weighted_node_features, axis=[-1])
        weighted_node_features = tf.linalg.diag(weighted_node_features)
        weighted_node_features = ops.modal_dot(
            incidence, weighted_node_features, transpose_a=True
        )
        weighted_node_features = ops.modal_dot(weighted_node_features, incidence)

        edge_adjacency = weighted_node_features * laplacian
        output = ops.modal_dot(edge_adjacency, edge_features)
        output = ops.modal_dot(output, self.edge_kernel)

        return self._bias_and_activation(output, bias_weights=self.edge_bias, mask=mask)

    def call(self, inputs, mask=None):
        node_features = self._propagate_nodes(inputs, mask=mask)
        edge_features = self._propagate_edges(inputs, mask=mask)

        return node_features, edge_features

    @property
    def config(self):
        # Get configuration for sub-components.
        node_reg = tf.keras.regularizers.serialize(self.__node_regularizer)
        node_init = tf.keras.initializers.serialize(self.__node_initializer)
        node_constraint = tf.keras.constraints.serialize(self.__node_constraint)

        edge_reg = tf.keras.regularizers.serialize(self.__edge_regularizer)
        edge_init = tf.keras.initializers.serialize(self.__edge_initializer)
        edge_constraint = tf.keras.constraints.serialize(self.__edge_constraint)

        return dict(
            node_channels=self.node_channels,
            edge_channels=self.edge_channels,
            node_regularizer=node_reg,
            node_initializer=node_init,
            node_constraint=node_constraint,
            edge_regularizer=edge_reg,
            edge_initializer=edge_init,
            edge_constraint=edge_constraint,
        )

    @staticmethod
    def preprocess(adjacency):
        laplacian = gcn_filter(adjacency)
        incidence = incidence_matrix(adjacency)
        edge_laplacian = gcn_filter(line_graph(incidence).numpy())

        return laplacian, edge_laplacian, incidence



from spektral.utils import label_to_one_hot, one_hot


class OneHotLabels:
    """
    One-hot encodes the graph labels along the innermost dimension (also if they
    are simple scalars).

    Either `depth` or `labels` must be passed as argument.

    **Arguments**

    - `depth`: int, the size of the one-hot vector (labels are intended as
    indices for a vector of size `depth`);
    - `labels`: list or tuple, the possible values that the labels can take
    (labels are one-hot encoded according to where they are found in`labels`).
    """

    def __init__(self, depth=None, labels=None):
        self.depth = depth
        self.labels = labels
        if self.depth is None and self.labels is None:
            raise ValueError("Must specify either depth or labels.")

    def __call__(self, graph):
        if self.labels is not None:
            graph.y = label_to_one_hot(graph.y, self.labels)
        else:
            graph.y = one_hot(graph.y, self.depth)

        return graph



import numpy as np
import scipy.sparse as sp
from sklearn.neighbors import kneighbors_graph
from tensorflow.keras.datasets import mnist as m

from spektral.data import Dataset, Graph

MNIST_SIZE = 28


class MNIST(Dataset):
    """
    The MNIST images used as node features for a grid graph, as described by
    [Defferrard et al. (2016)](https://arxiv.org/abs/1606.09375).

    This dataset is a graph signal classification task, where graphs are
    represented in mixed mode: one adjacency matrix, many instances of node
    features.

    For efficiency, the adjacency matrix is stored in a special attribute of the
    dataset and the Graphs only contain the node features.
    You can access the adjacency matrix via the `a` attribute.

    The node features of each graph are the MNIST digits vectorized and rescaled
    to [0, 1].
    Two nodes are connected if they are neighbours on the grid.
    Labels represent the MNIST class associated to each sample.

    **Note:** the last 10000 samples are the default test set of the MNIST
    dataset.

    **Arguments**

    - `p_flip`: if >0, then edges are randomly flipped from 0 to 1 or vice versa
    with that probability.
    - `k`: number of neighbours of each node.
    """

    def __init__(self, p_flip=0.0, k=8, **kwargs):
        self.a = None
        self.k = k
        self.p_flip = p_flip
        super().__init__(**kwargs)

    def read(self):
        self.a = _mnist_grid_graph(self.k)
        self.a = _flip_random_edges(self.a, self.p_flip)

        (x_train, y_train), (x_test, y_test) = m.load_data()
        x = np.vstack((x_train, x_test))
        x = x / 255.0
        y = np.concatenate((y_train, y_test), 0)
        x = x.reshape(-1, MNIST_SIZE**2, 1)

        return [Graph(x=x_, y=y_) for x_, y_ in zip(x, y)]


def _grid_coordinates(side):
    M = side**2
    x = np.linspace(0, 1, side, dtype=np.float32)
    y = np.linspace(0, 1, side, dtype=np.float32)
    xx, yy = np.meshgrid(x, y)
    z = np.empty((M, 2), np.float32)
    z[:, 0] = xx.reshape(M)
    z[:, 1] = yy.reshape(M)
    return z


def _get_adj_from_data(X, k, **kwargs):
    A = kneighbors_graph(X, k, **kwargs).toarray()
    A = sp.csr_matrix(np.maximum(A, A.T))

    return A


def _mnist_grid_graph(k):
    X = _grid_coordinates(MNIST_SIZE)
    A = _get_adj_from_data(
        X, k, mode="connectivity", metric="euclidean", include_self=False
    )

    return A


def _flip_random_edges(A, p_swap):
    if not A.shape[0] == A.shape[1]:
        raise ValueError("A must be a square matrix.")
    dtype = A.dtype
    A = sp.lil_matrix(A).astype(np.bool)
    n_elem = A.shape[0] ** 2
    n_elem_to_flip = round(p_swap * n_elem)
    unique_idx = np.random.choice(n_elem, replace=False, size=n_elem_to_flip)
    row_idx = unique_idx // A.shape[0]
    col_idx = unique_idx % A.shape[0]
    idxs = np.stack((row_idx, col_idx)).T
    for i in idxs:
        i = tuple(i)
        A[i] = np.logical_not(A[i])
    A = A.tocsr().astype(dtype)
    A.eliminate_zeros()
    return A



import copy
import warnings
from functools import partial

import numpy as np
import tensorflow as tf
from scipy import linalg
from scipy import sparse as sp
from scipy.sparse.linalg import ArpackNoConvergence


def degree_matrix(A):
    """
    Computes the degree matrix of the given adjacency matrix.
    :param A: rank 2 array or sparse matrix.
    :return: if A is a dense array, a dense array; if A is sparse, a sparse
    matrix in DIA format.
    """
    degrees = np.array(A.sum(1)).flatten()
    if sp.issparse(A):
        D = sp.diags(degrees)
    else:
        D = np.diag(degrees)
    return D


def degree_power(A, k):
    r"""
    Computes \(\D^{k}\) from the given adjacency matrix. Useful for computing
    normalised Laplacian.
    :param A: rank 2 array or sparse matrix.
    :param k: exponent to which elevate the degree matrix.
    :return: if A is a dense array, a dense array; if A is sparse, a sparse
    matrix in DIA format.
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        degrees = np.power(np.array(A.sum(1)), k).ravel()
    degrees[np.isinf(degrees)] = 0.0
    if sp.issparse(A):
        D = sp.diags(degrees)
    else:
        D = np.diag(degrees)
    return D


def normalized_adjacency(A, symmetric=True):
    r"""
    Normalizes the given adjacency matrix using the degree matrix as either
    \(\D^{-1}\A\) or \(\D^{-1/2}\A\D^{-1/2}\) (symmetric normalization).
    :param A: rank 2 array or sparse matrix;
    :param symmetric: boolean, compute symmetric normalization;
    :return: the normalized adjacency matrix.
    """
    if symmetric:
        normalized_D = degree_power(A, -0.5)
        return normalized_D.dot(A).dot(normalized_D)
    else:
        normalized_D = degree_power(A, -1.0)
        return normalized_D.dot(A)


def laplacian(A):
    r"""
    Computes the Laplacian of the given adjacency matrix as \(\D - \A\).
    :param A: rank 2 array or sparse matrix;
    :return: the Laplacian.
    """
    return degree_matrix(A) - A


def normalized_laplacian(A, symmetric=True):
    r"""
    Computes a  normalized Laplacian of the given adjacency matrix as
    \(\I - \D^{-1}\A\) or \(\I - \D^{-1/2}\A\D^{-1/2}\) (symmetric normalization).
    :param A: rank 2 array or sparse matrix;
    :param symmetric: boolean, compute symmetric normalization;
    :return: the normalized Laplacian.
    """
    if sp.issparse(A):
        I = sp.eye(A.shape[-1], dtype=A.dtype)
    else:
        I = np.eye(A.shape[-1], dtype=A.dtype)
    normalized_adj = normalized_adjacency(A, symmetric=symmetric)
    return I - normalized_adj


def rescale_laplacian(L, lmax=None):
    """
    Rescales the Laplacian eigenvalues in [-1,1], using lmax as largest eigenvalue.
    :param L: rank 2 array or sparse matrix;
    :param lmax: if None, compute largest eigenvalue with scipy.linalg.eisgh.
    If the eigendecomposition fails, lmax is set to 2 automatically.
    If scalar, use this value as largest eigenvalue when rescaling.
    :return:
    """
    if lmax is None:
        try:
            if sp.issparse(L):
                lmax = sp.linalg.eigsh(L, 1, which="LM", return_eigenvectors=False)[0]
            else:
                n = L.shape[-1]
                lmax = linalg.eigh(L, eigvals_only=True, eigvals=[n - 2, n - 1])[-1]
        except ArpackNoConvergence:
            lmax = 2
    if sp.issparse(L):
        I = sp.eye(L.shape[-1], dtype=L.dtype)
    else:
        I = np.eye(L.shape[-1], dtype=L.dtype)
    L_scaled = (2.0 / lmax) * L - I
    return L_scaled


def gcn_filter(A, symmetric=True):
    r"""
    Computes the graph filter described in
    [Kipf & Welling (2017)](https://arxiv.org/abs/1609.02907).
    :param A: array or sparse matrix with rank 2 or 3;
    :param symmetric: boolean, whether to normalize the matrix as
    \(\D^{-\frac{1}{2}}\A\D^{-\frac{1}{2}}\) or as \(\D^{-1}\A\);
    :return: array or sparse matrix with rank 2 or 3, same as A;
    """
    out = copy.deepcopy(A)
    if isinstance(A, list) or (isinstance(A, np.ndarray) and A.ndim == 3):
        for i in range(len(A)):
            out[i] = A[i]
            out[i][np.diag_indices_from(out[i])] += 1
            out[i] = normalized_adjacency(out[i], symmetric=symmetric)
    else:
        if hasattr(out, "tocsr"):
            out = out.tocsr()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            out[np.diag_indices_from(out)] += 1
        out = normalized_adjacency(out, symmetric=symmetric)

    if sp.issparse(out):
        out.sort_indices()
    return out


def _triangular_adjacency(adjacency):
    """
    Gets the triangular version of the adjacency matrix, removing redundant
    values.

    :param adjacency: The full adjacency matrix, with shape
        ([batch], n_nodes, n_nodes).
    :return: The upper triangle of the adjacency matrix, with the lower
        triangle set to zero.
    """
    return tf.linalg.band_part(adjacency, 0, -1)


def _incidence_matrix_single(triangular_adjacency, *, num_edges):
    """
    Creates the corresponding incidence matrix for a graph with a particular
    adjacency matrix.

    :param triangular_adjacency: The binary adjacency matrix. Should have shape
        (n_nodes, n_nodes), and be triangular.
    :param num_edges: The number of edges to use in the output. Should be large
        enough to accommodate all the edges in the adjacency matrix.

    :return: The computed incidence matrix. It will have a shape of
        (n_nodes, n_edges).
    """
    # The adjacency matrix should be sparse, so get the indices of the edges.
    connected_node_indices = tf.where(triangular_adjacency)

    # Match each edge with one of the nodes connected by that edge. We refer
    # to the two nodes connected by each edge as "right" and "left",
    # for convenience.
    edge_indices = tf.range(connected_node_indices.shape[0], dtype=tf.int64)
    edges_with_left_nodes = tf.stack(
        [connected_node_indices[:, 0], edge_indices], axis=1
    )
    edges_with_right_nodes = tf.stack(
        [connected_node_indices[:, 1], edge_indices], axis=1
    )

    # We now have all the points that should go in the sparse binary
    # transformation matrix.
    edge_indicators = tf.ones_like(edge_indices, dtype=tf.float32)
    num_nodes = tf.cast(tf.shape(triangular_adjacency)[0], tf.int64)
    output_shape = tf.stack([num_nodes, num_edges])
    left_sparse = tf.SparseTensor(
        indices=edges_with_left_nodes, values=edge_indicators, dense_shape=output_shape
    )
    left_sparse = tf.sparse.reorder(left_sparse)
    right_sparse = tf.SparseTensor(
        indices=edges_with_right_nodes, values=edge_indicators, dense_shape=output_shape
    )
    right_sparse = tf.sparse.reorder(right_sparse)
    # Combine the matrices for the left and right nodes.
    combined_sparse = tf.sparse.maximum(left_sparse, right_sparse)

    return tf.sparse.to_dense(combined_sparse)


def incidence_matrix(adjacency):
    """
    Creates the corresponding incidence matrices for graphs with particular
    adjacency matrices.

    :param adjacency: The binary adjacency matrices. Should have shape
        ([batch], n_nodes, n_nodes).
    :return: The computed incidence matrices. It will have a shape of
        ([batch], n_nodes, n_edges).
    """
    adjacency = tf.convert_to_tensor(adjacency, dtype=tf.float32)
    added_batch = False
    if tf.size(tf.shape(adjacency)) == 2:
        # Add the extra batch dimension if needed.
        adjacency = tf.expand_dims(adjacency, axis=0)
        added_batch = True

    # Compute the maximum number of edges. We will pad everything in the
    # batch to this dimension.
    adjacency_upper = _triangular_adjacency(adjacency)
    num_edges = tf.math.count_nonzero(adjacency_upper, axis=(1, 2))
    max_num_edges = tf.reduce_max(num_edges)

    # Compute all the transformation matrices.
    make_single_matrix = partial(_incidence_matrix_single, num_edges=max_num_edges)
    transformation_matrices = tf.map_fn(
        make_single_matrix,
        adjacency_upper,
        fn_output_signature=tf.TensorSpec(shape=[None, None], dtype=tf.float32),
    )

    if added_batch:
        # Remove the extra batch dimension before returning.
        transformation_matrices = transformation_matrices[0]
    return transformation_matrices


def line_graph(incidence):
    """
    Creates the line graph adjacency matrices for graphs with particular
    incidence matrices.
    :param incidence: The incidence matrices. Should have shape
        ([batch], n_nodes, n_edges).
    :return: The computed line graph adjacency matrices. It will have a shape
        of ([batch], n_edges, n_edges).
    """
    incidence = tf.convert_to_tensor(incidence, dtype=tf.float32)

    incidence_t = tf.linalg.matrix_transpose(incidence)
    incidence_sq = tf.matmul(incidence_t, incidence)

    num_rows = tf.shape(incidence_sq)[-2]
    identity = tf.eye(num_rows)
    return incidence_sq - identity * 2


def chebyshev_polynomial(X, k):
    """
    Calculates Chebyshev polynomials of X, up to order k.
    :param X: rank 2 array or sparse matrix;
    :param k: the order up to which compute the polynomials,
    :return: a list of k + 1 arrays or sparse matrices with one element for each
    degree of the polynomial.
    """
    T_k = list()
    if sp.issparse(X):
        T_k.append(sp.eye(X.shape[0], dtype=X.dtype).tocsr())
    else:
        T_k.append(np.eye(X.shape[0], dtype=X.dtype))
    T_k.append(X)

    def chebyshev_recurrence(T_k_minus_one, T_k_minus_two, X):
        if sp.issparse(X):
            X_ = sp.csr_matrix(X, copy=True)
        else:
            X_ = np.copy(X)
        return 2 * X_.dot(T_k_minus_one) - T_k_minus_two

    for _ in range(2, k + 1):
        T_k.append(chebyshev_recurrence(T_k[-1], T_k[-2], X))

    return T_k


def chebyshev_filter(A, k, symmetric=True):
    r"""
    Computes the Chebyshev filter from the given adjacency matrix, as described
    in [Defferrard et at. (2016)](https://arxiv.org/abs/1606.09375).
    :param A: rank 2 array or sparse matrix;
    :param k: integer, the order of the Chebyshev polynomial;
    :param symmetric: boolean, whether to normalize the matrix as
    \(\D^{-\frac{1}{2}}\A\D^{-\frac{1}{2}}\) or as \(\D^{-1}\A\);
    :return: a list of k + 1 arrays or sparse matrices with one element for each
    degree of the polynomial.
    """
    normalized_adj = normalized_adjacency(A, symmetric)
    if sp.issparse(A):
        I = sp.eye(A.shape[0], dtype=A.dtype)
    else:
        I = np.eye(A.shape[0], dtype=A.dtype)
    L = I - normalized_adj  # Compute Laplacian

    # Rescale Laplacian
    L_scaled = rescale_laplacian(L)

    # Compute Chebyshev polynomial approximation
    T_k = chebyshev_polynomial(L_scaled, k)

    # Sort indices
    if sp.issparse(T_k[0]):
        for i in range(len(T_k)):
            T_k[i].sort_indices()

    return T_k


def add_self_loops(a, value=1):
    """
    Sets the inner diagonals of `a` to `value`.
    :param a: a np.array or scipy.sparse matrix, the innermost two dimensions
    must be equal.
    :param value: value to set the diagonals to.
    :return: a np.array or scipy.sparse matrix with the same shape as `a`.
    """
    a = a.copy()
    if len(a.shape) < 2:
        raise ValueError("a must have at least rank 2")
    n = a.shape[-1]
    if n != a.shape[-2]:
        raise ValueError(
            "Innermost two dimensions must be equal. Got {}".format(a.shape)
        )
    if sp.issparse(a):
        a = a.tolil()
        a.setdiag(value)
        return a.tocsr()
    else:
        idx = np.arange(n)
        a[..., idx, idx] = value
        return a



import errno
import os
import os.path as osp
import ssl
import sys
import urllib

import numpy as np
import scipy.sparse as sp

from spektral.data import Dataset, Graph
from spektral.datasets.citation import _preprocess_features
from spektral.datasets.utils import DATASET_FOLDER
from spektral.utils import label_to_one_hot


class DBLP(Dataset):
    """
    A subset of the DBLP computer science bibliography website,
    as collected in the
    [Fu et al. (2020)](https://arxiv.org/abs/2002.01680) paper.

    **Arguments**

    - `normalize_x`: if True, normalize the features.
    - `dtype`: numpy dtype of graph data.
    """

    def __init__(self, normalize_x=False, dtype=np.float32, **kwargs):
        self.dtype = dtype
        self.normalize_x = normalize_x
        super().__init__(**kwargs)

    url = "https://github.com/abojchevski/graph2gauss/raw/master/data/dblp.npz"

    @property
    def path(self):
        return osp.join(DATASET_FOLDER, "Citation", "dblp")

    def download(self):
        print("Downloading DBLP dataset.")
        _download_url(self.url, self.path)

    def read(self):
        f = np.load(osp.join(self.path, "dblp.npz"))

        x = sp.csr_matrix(
            (f["attr_data"], f["attr_indices"], f["attr_indptr"]), f["attr_shape"]
        ).toarray()
        x[x > 0] = 1

        if self.normalize_x:
            print("Pre-processing node features")
            x = _preprocess_features(x)

        a = sp.csr_matrix(
            (f["adj_data"], f["adj_indices"], f["adj_indptr"]), f["adj_shape"]
        )  # .tocoo()

        y = f["labels"]
        y = label_to_one_hot(y, np.unique(y))

        return [
            Graph(
                x=x.astype(self.dtype),
                a=a.astype(self.dtype),
                y=y.astype(self.dtype),
            )
        ]


def _download_url(url, folder, log=False):
    r"""Downloads the content of an URL to a specific folder.
    Args:
        url (string): The url.
        folder (string): The folder.
        log (bool, optional): If :obj:`False`, will not print anything to the
            console. (default: :obj:`True`)
    """

    filename = url.rpartition("/")[2]
    filename = filename if filename[0] == "?" else filename.split("?")[0]
    path = osp.join(folder, filename)

    if osp.exists(path):  # pragma: no cover
        if log:
            print(f"Using existing file {filename}", file=sys.stderr)
        return path

    if log:
        print(f"Downloading {url}", file=sys.stderr)

    try:
        os.makedirs(osp.expanduser(osp.normpath(folder)), exist_ok=True)
    except OSError as e:
        if e.errno != errno.EEXIST and osp.isdir(path):
            raise e

    context = ssl._create_unverified_context()
    data = urllib.request.urlopen(url, context=context)

    with open(path, "wb") as f:
        f.write(data.read())

    return path



import tensorflow as tf
from tensorflow.python.ops import gen_sparse_ops

from . import ops


def add_self_loops(a, fill=1.0):
    """
    Adds self-loops to the given adjacency matrix. Self-loops are added only for
    those node that don't have a self-loop already, and are assigned a weight
    of `fill`.
    :param a: a square SparseTensor.
    :param fill: the fill value for the new self-loops. It will be cast to the
    dtype of `a`.
    :return: a SparseTensor with the same shape as the input.
    """
    indices = a.indices
    values = a.values
    N = tf.shape(a, out_type=indices.dtype)[0]

    mask_od = indices[:, 0] != indices[:, 1]
    mask_sl = ~mask_od
    mask_od.set_shape([None])  # For compatibility with TF 2.2
    mask_sl.set_shape([None])

    indices_od = indices[mask_od]
    indices_sl = indices[mask_sl]

    values_sl = tf.fill((N,), tf.cast(fill, values.dtype))
    values_sl = tf.tensor_scatter_nd_update(
        values_sl, indices_sl[:, 0:1], values[mask_sl]
    )

    indices_sl = tf.range(N, dtype=indices.dtype)[:, None]
    indices_sl = tf.repeat(indices_sl, 2, -1)
    indices = tf.concat((indices_od, indices_sl), 0)

    values_od = values[mask_od]
    values = tf.concat((values_od, values_sl), 0)

    out = tf.SparseTensor(indices, values, (N, N))

    return tf.sparse.reorder(out)


def add_self_loops_indices(indices, n_nodes=None):
    """
    Given the indices of a square SparseTensor, adds the diagonal entries (i, i)
    and returns the reordered indices.
    :param indices: Tensor of rank 2, the indices to a SparseTensor.
    :param n_nodes: the size of the n_nodes x n_nodes SparseTensor indexed by
    the indices. If `None`, n_nodes is calculated as the maximum entry in the
    indices plus 1.
    :return: Tensor of rank 2, the indices to a SparseTensor.
    """
    n_nodes = tf.reduce_max(indices) + 1 if n_nodes is None else n_nodes
    row, col = indices[..., 0], indices[..., 1]
    mask = tf.ensure_shape(row != col, row.shape)
    sl_indices = tf.range(n_nodes, dtype=row.dtype)[:, None]
    sl_indices = tf.repeat(sl_indices, 2, -1)
    indices = tf.concat((indices[mask], sl_indices), 0)
    dummy_values = tf.ones_like(indices[:, 0])
    indices, _ = gen_sparse_ops.sparse_reorder(
        indices, dummy_values, (n_nodes, n_nodes)
    )
    return indices


def _square_size(dense_shape):
    dense_shape = tf.unstack(dense_shape)
    size = dense_shape[0]
    for d in dense_shape[1:]:
        tf.debugging.assert_equal(size, d)
    return size


def _indices_to_inverse_map(indices, size):
    """
    Compute inverse indices of a gather.
    :param indices: Tensor, forward indices, rank 1
    :param size: Tensor, size of pre-gathered input, rank 0
    :return: Tensor, inverse indices, shape [size]. Zero values everywhere
    except at indices.
    """
    indices = tf.cast(indices, tf.int64)
    size = tf.cast(size, tf.int64)
    return tf.scatter_nd(
        tf.expand_dims(indices, axis=-1),
        tf.range(tf.shape(indices, out_type=tf.int64)[0]),
        tf.expand_dims(size, axis=-1),
    )


def _boolean_mask_sparse(a, mask, axis, inverse_map, out_size):
    """
    SparseTensor equivalent to tf.boolean_mask.
    :param a: SparseTensor of rank k and nnz non-zeros.
    :param mask: rank-1 bool Tensor.
    :param axis: int, axis on which to mask. Must be in [-k, k).
    :param out_size: number of true entires in mask. Computed if not given.
    :return masked_a: SparseTensor masked along the given axis.
    :return values_mask: bool Tensor indicating surviving edges, shape [nnz].
    """
    mask = tf.convert_to_tensor(mask)
    values_mask = tf.gather(mask, a.indices[:, axis], axis=0)
    dense_shape = tf.tensor_scatter_nd_update(a.dense_shape, [[axis]], [out_size])
    indices = tf.boolean_mask(a.indices, values_mask)
    indices = tf.unstack(indices, axis=-1)
    indices[axis] = tf.gather(inverse_map, indices[axis])
    indices = tf.stack(indices, axis=-1)
    a = tf.SparseTensor(
        indices,
        tf.boolean_mask(a.values, values_mask),
        dense_shape,
    )
    return (a, values_mask)


def _boolean_mask_sparse_square(a, mask, inverse_map, out_size):
    """
    Apply boolean_mask to every axis of a SparseTensor.
    :param a: SparseTensor with uniform dimensions and nnz non-zeros.
    :param mask: boolean mask.
    :param inverse_map: Tensor of new indices, shape [nnz]. Computed if None.
    :out_size: number of True values in mask. Computed if None.
    :return a: SparseTensor with uniform dimensions.
    :return values_mask: bool Tensor of shape [nnz] indicating valid edges.
    """
    mask = tf.convert_to_tensor(mask)
    values_mask = tf.reduce_all(tf.gather(mask, a.indices, axis=0), axis=-1)
    dense_shape = [out_size] * a.shape.ndims
    indices = tf.boolean_mask(a.indices, values_mask)
    indices = tf.gather(inverse_map, indices)
    a = tf.SparseTensor(indices, tf.boolean_mask(a.values, values_mask), dense_shape)
    return (a, values_mask)


def boolean_mask_sparse(a, mask, axis=0):
    """
    SparseTensor equivalent to tf.boolean_mask.
    :param a: SparseTensor of rank k and nnz non-zeros.
    :param mask: rank-1 bool Tensor.
    :param axis: int, axis on which to mask. Must be in [-k, k).
    :return masked_a: SparseTensor masked along the given axis.
    :return values_mask: bool Tensor indicating surviving values, shape [nnz].
    """
    i = tf.squeeze(tf.where(mask), axis=1)
    out_size = tf.math.count_nonzero(mask)
    in_size = a.dense_shape[axis]
    inverse_map = _indices_to_inverse_map(i, in_size)
    return _boolean_mask_sparse(
        a, mask, axis=axis, inverse_map=inverse_map, out_size=out_size
    )


def boolean_mask_sparse_square(a, mask):
    """
    Apply mask to every axis of SparseTensor a.
    :param a: SparseTensor, square, nnz non-zeros.
    :param mask: boolean mask with size equal to each dimension of a.
    :return masked_a: SparseTensor
    :return values_mask: bool tensor of shape [nnz] indicating valid values.
    """
    i = tf.squeeze(tf.where(mask), axis=-1)
    out_size = tf.size(i)
    in_size = _square_size(a.dense_shape)
    inverse_map = _indices_to_inverse_map(i, in_size)
    return _boolean_mask_sparse_square(
        a, mask, inverse_map=inverse_map, out_size=out_size
    )


def gather_sparse(a, indices, axis=0, mask=None):
    """
    SparseTensor equivalent to tf.gather, assuming indices are sorted.
    :param a: SparseTensor of rank k and nnz non-zeros.
    :param indices: rank-1 int Tensor, rows or columns to keep.
    :param axis: int axis to apply gather to.
    :param mask: boolean mask corresponding to indices. Computed if not provided.
    :return gathered_a: SparseTensor masked along the given axis.
    :return values_mask: bool Tensor indicating surviving values, shape [nnz].
    """
    in_size = _square_size(a.dense_shape)
    out_size = tf.size(indices)
    if mask is None:
        mask = ops.indices_to_mask(indices, in_size)
    inverse_map = _indices_to_inverse_map(indices, in_size)
    return _boolean_mask_sparse(
        a, mask, axis=axis, inverse_map=inverse_map, out_size=out_size
    )


def gather_sparse_square(a, indices, mask=None):
    """
    Gather on every axis of a SparseTensor.
    :param a: SparseTensor of rank k and nnz non-zeros.
    :param indices: rank-1 int Tensor, rows and columns to keep.
    :param mask: boolean mask corresponding to indices. Computed if not provided.
    :return gathered_a: SparseTensor of the gathered input.
    :return values_mask: bool Tensor indicating surviving values, shape [nnz].
    """
    in_size = _square_size(a.dense_shape)
    out_size = tf.size(indices)
    if mask is None:
        mask = ops.indices_to_mask(indices, in_size)
    inverse_map = _indices_to_inverse_map(indices, in_size)
    return _boolean_mask_sparse_square(
        a, mask, inverse_map=inverse_map, out_size=out_size
    )
